#include "Ocean Optics XOP.h"

// Runtime param structure for ListAndOpenSpectrometers operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct ListAndOpenSpectrometersRuntimeParams {
	// Flag parameters.
	
	// Parameters for /IP flag group.
	int IPFlagEncountered;
	waveHndl IPTextWave;
	int IPFlagParamsSet[1];
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct ListAndOpenSpectrometersRuntimeParams ListAndOpenSpectrometersRuntimeParams;
typedef struct ListAndOpenSpectrometersRuntimeParams* ListAndOpenSpectrometersRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.

// Runtime param structure for CloseSpectrometers operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct CloseSpectrometersRuntimeParams {
	// Flag parameters.
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct CloseSpectrometersRuntimeParams CloseSpectrometersRuntimeParams;
typedef struct CloseSpectrometersRuntimeParams* CloseSpectrometersRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.

// Runtime param structure for SetupSpectrometer operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct SetupSpectrometerRuntimeParams {
	// Flag parameters.
	
	// Parameters for /ID flag group.
	int IDFlagEncountered;
	double spectrometerID;
	int IDFlagParamsSet[1];
	
	// Parameters for /AVG flag group.
	int AVGFlagEncountered;
	double scansToAverage;
	int AVGFlagParamsSet[1];
	
	// Parameters for /BXCR flag group.
	int BXCRFlagEncountered;
	double boxcarWidth;
	int BXCRFlagParamsSet[1];
	
	// Parameters for /ASTB flag group.
	int ASTBFlagEncountered;
	double autoToggleLamp;
	int ASTBFlagParamsSet[1];
	
	// Parameters for /INTG flag group.
	int INTGFlagEncountered;
	double integrationTime;
	int INTGFlagParamsSet[1];
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct SetupSpectrometerRuntimeParams SetupSpectrometerRuntimeParams;
typedef struct SetupSpectrometerRuntimeParams* SetupSpectrometerRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.

// Runtime param structure for GetSpectrometerSetup operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct GetSpectrometerSetupRuntimeParams {
	// Flag parameters.
	
	// Parameters for /ID flag group.
	int IDFlagEncountered;
	double spectrometerID;
	int IDFlagParamsSet[1];
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct GetSpectrometerSetupRuntimeParams GetSpectrometerSetupRuntimeParams;
typedef struct GetSpectrometerSetupRuntimeParams* GetSpectrometerSetupRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.

// Runtime param structure for AcquireSpectrometerSpectrum operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct AcquireSpectrometerSpectrumRuntimeParams {
	// Flag parameters.
	
	// Parameters for /ID flag group.
	int IDFlagEncountered;
	double spectrometerID;
	int IDFlagParamsSet[1];
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct AcquireSpectrometerSpectrumRuntimeParams AcquireSpectrometerSpectrumRuntimeParams;
typedef struct AcquireSpectrometerSpectrumRuntimeParams* AcquireSpectrometerSpectrumRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.

// Runtime param structure for CloseAllSpectrometers operation.
#include "XOPStructureAlignmentTwoByte.h"	// All structures passed to Igor are two-byte aligned.
struct CloseAllSpectrometersRuntimeParams {
	// Flag parameters.
	
	// Main parameters.
	
	// These are postamble fields that Igor sets.
	int calledFromFunction;					// 1 if called from a user function, 0 otherwise.
	int calledFromMacro;					// 1 if called from a macro, 0 otherwise.
};
typedef struct CloseAllSpectrometersRuntimeParams CloseAllSpectrometersRuntimeParams;
typedef struct CloseAllSpectrometersRuntimeParams* CloseAllSpectrometersRuntimeParamsPtr;
#include "XOPStructureAlignmentReset.h"		// Reset structure alignment to default.



static int ExecuteListAndOpenSpectrometers(ListAndOpenSpectrometersRuntimeParamsPtr p) {
	int err = 0;
	int nSpectrometers = 0;
	JString errMessage;
	JString spectrometerName;
	
	// if required then initialize the wrapper object
	if (globalWrapperPtr == NULL) {
		globalWrapperPtr = new Wrapper();
		// get all spectrometers directly connected to the computer
		nSpectrometers = globalWrapperPtr->openAllSpectrometers();
		if (nSpectrometers == -1) {	// error
			errMessage = globalWrapperPtr->getLastException();
			XOPNotice(errMessage.getASCII());
			XOPNotice("\r");
			delete globalWrapperPtr;
			globalWrapperPtr = NULL;
			return NOMEM;	// placeholder error
		}
	} else {	// the wrapper object already exists, which means the spectrometers have already been opened
		nSpectrometers = globalWrapperPtr->getNumberOfSpectrometersFound();
	}
	
	if (p->IPFlagEncountered) {
		// Parameter: p->IPTextWave (test for NULL handle before using)
		if (p->IPTextWave == NULL)
			return NOWAV;
		// TODO: open network spectrometers here
	}
	
	// setup some default values
	for (size_t i = 0; i < nSpectrometers; ++i) {
		globalWrapperPtr->setBoxcarWidth(i,0);
		globalWrapperPtr->setCorrectForDetectorNonlinearity(i, true);
		globalWrapperPtr->setCorrectForElectricalDark(i, true);
		globalWrapperPtr->setScansToAverage(i, 1);
		globalWrapperPtr->setIntegrationTime(i, 1.0e4);	// default integration time 10 ms
	}
	
	// return a text wave to Igor containing the names of the spectrometers
	long dimensionSizes[MAX_DIMENSIONS+1];
	long indices[MAX_DIMENSIONS];
	waveHndl spectrometerListWave;
	Handle h = NewHandle(512);
	
	dimensionSizes[0] = nSpectrometers;
	dimensionSizes[1] = 0;
	
	err = MDMakeWave(&spectrometerListWave, "W_OpenSpectrometers", NULL, dimensionSizes, TEXT_WAVE_TYPE, 1);
	if (err != 0) {
		return err;
	}
	
	indices[1] = 0;
	for (size_t i = 0; i < nSpectrometers; ++i) {
		spectrometerName = globalWrapperPtr->getName(i);
		
		err = PutCStringInHandle(spectrometerName.getASCII(), h);
		if (err != 0)
			return err;
		
		indices[0] = i;
		err = MDSetTextWavePointValue(spectrometerListWave, indices, h);
		if (err != 0)
			return err;
	}
	
	return err;
}

static int ExecuteCloseSpectrometers(CloseSpectrometersRuntimeParamsPtr p) {
	int err = 0;
	
	// do we have any cameras open?
	if (globalWrapperPtr == NULL) {	// no cameras open, do nothing
		return 0;
	} else {
		globalWrapperPtr->closeAllSpectrometers();
		globalWrapperPtr = NULL;
	}
	
	return err;
}

static int ExecuteSetupSpectrometer(SetupSpectrometerRuntimeParamsPtr p) {
	int err = 0;
	size_t spectrometerID, scansToAverage, boxcarWidth;
	size_t autoToggleLamp;
	double integrationTime;
	
	// check if the wrapper class has been initialized
	if (globalWrapperPtr == NULL) {
		XOPNotice("No spectrometers have been initialized\r");
		return NOMEM;
	}
	
	// Flag parameters.
	
	if (p->IDFlagEncountered) {
		// Parameter: p->spectrometerID
		spectrometerID = p->spectrometerID;
	} else {
		return EXPECT_POS_NUM;
	}

	// check if the required spectrometer exists
	if (spectrometerID >= globalWrapperPtr->getNumberOfSpectrometersFound()) {
		XOPNotice("Invalid spectrometerID\r");
		return NOMEM;
	}
	
	if (p->AVGFlagEncountered) {
		// Parameter: p->scansToAverage
		if (p->scansToAverage < 0)
			return EXPECT_POS_NUM;
		scansToAverage = (size_t)(p->scansToAverage + 0.5);
		globalWrapperPtr->setScansToAverage(spectrometerID, scansToAverage);
	}
	
	if (p->BXCRFlagEncountered) {
		// Parameter: p->boxcarWidth
		if (p->boxcarWidth < 0)
			return EXPECT_POS_NUM;
		boxcarWidth = (size_t)(p->boxcarWidth);
		globalWrapperPtr->setBoxcarWidth(spectrometerID, boxcarWidth);
	}
	
	if (p->ASTBFlagEncountered) {
		// Parameter: p->autoToggleLamp
		if (p->autoToggleLamp < 0)
			return EXPECT_POS_NUM;
		autoToggleLamp = (size_t)(p->autoToggleLamp + 0.5);
		if ((autoToggleLamp != 0) && (autoToggleLamp != 1)) {
			// the autotoggle is either on or off
			XOPNotice("The /ASTB flag accepts only 0 or 1 as a parameter\r");
		}
		
		if (autoToggleLamp == 1) {
			globalWrapperPtr->setStrobeEnable(spectrometerID ,true);
			globalWrapperPtr->setAutoToggleStrobeLampEnable(spectrometerID, true);
		} else {
			globalWrapperPtr->setAutoToggleStrobeLampEnable(spectrometerID, false);
		}
	}
	
	if (p->INTGFlagEncountered) {
		// Parameter: p->integrationTime
		if (p->integrationTime < 0)
			return EXPECT_POS_NUM;
		
		integrationTime = p->integrationTime;
		globalWrapperPtr->setIntegrationTime(spectrometerID, integrationTime * 1.0e6);
	}

	return err;
}

static int ExecuteGetSpectrometerSetup(GetSpectrometerSetupRuntimeParamsPtr p) {
	int err = 0;
	int spectrometerID, scansToAverage, boxcarWidth;
	int autoToggleLamp;
	JString jstring;
	size_t integrationTime;
	
	// check if the wrapper class has been initialized
	if (globalWrapperPtr == NULL) {
		XOPNotice("No spectrometers have been initialized\r");
		return NOMEM;
	}
	
	// Flag parameters.
	
	if (p->IDFlagEncountered) {
		// Parameter: p->spectrometerID
		if (p->spectrometerID < 0)
			return EXPECT_POS_NUM;
		
		spectrometerID = (size_t)(p->spectrometerID);
	} else {
		return EXPECT_POS_NUM;
	}

	// check if the required spectrometer exists
	if (spectrometerID >= globalWrapperPtr->getNumberOfSpectrometersFound()) {
		XOPNotice("Invalid spectrometerID\r");
		return NOMEM;
	}
	
	integrationTime = globalWrapperPtr->getIntegrationTime(spectrometerID);
	err = SetOperationNumVar("V_integrationTime", (double)integrationTime / 1.0e6);
	if (err != 0)
		return err;
	
	boxcarWidth = globalWrapperPtr->getBoxcarWidth(spectrometerID);
	err = SetOperationNumVar("V_boxCarWidth", boxcarWidth);
	if (err != 0)
		return err;
	
	scansToAverage = globalWrapperPtr->getScansToAverage(spectrometerID);
	err = SetOperationNumVar("V_scansToAverage", scansToAverage);
	if (err != 0)
		return err;
	
	jstring = globalWrapperPtr->getName(spectrometerID);
	err = SetOperationStrVar("S_spectrometerName", jstring.getASCII());
	if (err != 0)
		return err;
	
	jstring = globalWrapperPtr->getSerialNumber(spectrometerID);
	err = SetOperationStrVar("S_spectrometerSerialNumber", jstring.getASCII());
	if (err != 0)
		return err;
	
	return err;
}

static int ExecuteAcquireSpectrometerSpectrum(AcquireSpectrometerSpectrumRuntimeParamsPtr p) {
	int err = 0;
	size_t spectrometerID;
	
	// check if the wrapper class has been initialized
	if (globalWrapperPtr == NULL) {
		XOPNotice("No spectrometers have been initialized\r");
		return NOMEM;
	}
	
	// Flag parameters.
	
	if (p->IDFlagEncountered) {
		// Parameter: p->spectrometerID
		spectrometerID = (size_t)(p->spectrometerID + 0.5);
	} else {
		return EXPECT_POS_NUM;
	}

	// check if the required spectrometer exists
	if (spectrometerID >= globalWrapperPtr->getNumberOfSpectrometersFound()) {
		XOPNotice("Invalid spectrometerID\r");
		return NOMEM;
	}
	
	// get a spectrum
	DoubleArray spectrumArray, wavelengthArray;
	spectrumArray = globalWrapperPtr->getSpectrum(spectrometerID);
	wavelengthArray = globalWrapperPtr->getWavelengths(spectrometerID);
	
	if (spectrumArray.getLength() != wavelengthArray.getLength()) {
		XOPNotice("spectrumArray.getLength() != wavelengthArray.getLength()\r");
		return NOMEM;
	}
	
	size_t nValues = spectrumArray.getLength();
	double* spectrum = spectrumArray.getDoubleValues();
	double* wavelengths = wavelengthArray.getDoubleValues();
	long dimensionSizes[MAX_DIMENSIONS+1];
	waveHndl spectrumWave, wavelengthsWave;
	
	dimensionSizes[0] = nValues;
	dimensionSizes[1] = 0;
	
	// make some output waves
	err = MDMakeWave(&spectrumWave, "W_Spectrum", NULL, dimensionSizes, NT_FP64, 1);
	if (err != 0)
		return err;
	err = MDMakeWave(&wavelengthsWave, "W_Wavelengths", NULL, dimensionSizes, NT_FP64, 1);
	if (err != 0)
		return err;
	
	// copy the values to the wave
	err = MDStoreDPDataInNumericWave(spectrumWave, spectrum);
	if (err != 0)
		return err;
	
	err = MDStoreDPDataInNumericWave(wavelengthsWave, wavelengths);
	if (err != 0)
		return err;
	
	return err;
}

static int ExecuteCloseAllSpectrometers(CloseAllSpectrometersRuntimeParamsPtr p) {
	int err = 0;
	
	if (globalWrapperPtr != NULL) {
		globalWrapperPtr->closeAllSpectrometers();
		globalWrapperPtr = NULL;
	}
	
	return err;
}

static int RegisterListAndOpenSpectrometers(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the ListAndOpenSpectrometersRuntimeParams structure as well.
	cmdTemplate = "ListAndOpenSpectrometers /IP=wave:IPTextWave";
	runtimeNumVarList = "";
	runtimeStrVarList = "";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(ListAndOpenSpectrometersRuntimeParams), (void*)ExecuteListAndOpenSpectrometers, 0);
}

static int RegisterCloseSpectrometers(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the CloseSpectrometersRuntimeParams structure as well.
	cmdTemplate = "CloseSpectrometers";
	runtimeNumVarList = "";
	runtimeStrVarList = "";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(CloseSpectrometersRuntimeParams), (void*)ExecuteCloseSpectrometers, 0);
}

static int RegisterSetupSpectrometer(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the SetupSpectrometerRuntimeParams structure as well.
	cmdTemplate = "SetupSpectrometer /ID=number:spectrometerID /AVG=number:scansToAverage /BXCR=number:boxcarWidth /ASTB=number:autoToggleLamp /INTG=number:integrationTime";
	runtimeNumVarList = "";
	runtimeStrVarList = "";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(SetupSpectrometerRuntimeParams), (void*)ExecuteSetupSpectrometer, 0);
}

static int RegisterGetSpectrometerSetup(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the GetSpectrometerSetupRuntimeParams structure as well.
	cmdTemplate = "GetSpectrometerSetup /ID=number:spectrometerID";
	runtimeNumVarList = "V_integrationTime;V_scansToAverage;V_boxcarWidth;";
	runtimeStrVarList = "S_spectrometerName;S_spectrometerSerialNumber;";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(GetSpectrometerSetupRuntimeParams), (void*)ExecuteGetSpectrometerSetup, 0);
}

static int RegisterAcquireSpectrometerSpectrum(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the AcquireSpectrometerSpectrumRuntimeParams structure as well.
	cmdTemplate = "AcquireSpectrometerSpectrum /ID=number:spectrometerID";
	runtimeNumVarList = "";
	runtimeStrVarList = "";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(AcquireSpectrometerSpectrumRuntimeParams), (void*)ExecuteAcquireSpectrometerSpectrum, 0);
}

static int RegisterCloseAllSpectrometers(void) {
	const char* cmdTemplate;
	const char* runtimeNumVarList;
	const char* runtimeStrVarList;
	
	// NOTE: If you change this template, you must change the CloseAllSpectrometersRuntimeParams structure as well.
	cmdTemplate = "CloseAllSpectrometers";
	runtimeNumVarList = "";
	runtimeStrVarList = "";
	return RegisterOperation(cmdTemplate, runtimeNumVarList, runtimeStrVarList, sizeof(CloseAllSpectrometersRuntimeParams), (void*)ExecuteCloseAllSpectrometers, 0);
}


static int RegisterOperations(void)		// Register any operations with Igor.
{
	int result;

	// initialize the wrapper pointer to NULL
	globalWrapperPtr = NULL;
	
	if (result = RegisterListAndOpenSpectrometers())
		return result;
	
	if (result = RegisterCloseSpectrometers())
		return result;
	
	if (result = RegisterSetupSpectrometer())
		return result;
	
	if (result = RegisterAcquireSpectrometerSpectrum())
		return result;
	
	if (result = RegisterGetSpectrometerSetup())
		return result;
	
	if (result = RegisterCloseAllSpectrometers())
		return result;
	
	// There are no more operations added by this XOP.
		
	return 0;
}

/*	XOPEntry()

	This is the entry point from the host application to the XOP for all
	messages after the INIT message.
*/
static void XOPEntry(void) {	
	long result = 0;
	
	switch (GetXOPMessage()) {
		// We don't need to handle any messages for this XOP.
	}
	SetXOPResult(result);
}

/*	main(ioRecHandle)

	This is the initial entry point at which the host application calls XOP.
	The message sent by the host must be INIT.
	
	main does any necessary initialization and then sets the XOPEntry field of the
	ioRecHandle to the address to be called for future messages.
*/
HOST_IMPORT int main(IORecHandle ioRecHandle) {
	int result;
	
	XOPInit(ioRecHandle);							// Do standard XOP initialization.

	SetXOPEntry(XOPEntry);							// Set entry point for future calls.

	if (result = RegisterOperations()) {
		SetXOPResult(result);
	}
	else {
	SetXOPResult(0);
	}
}
